from collections import defaultdict
import datetime
import json
import os
import pathlib
import tarfile
from typing import Dict, Iterable, List, NamedTuple, Tuple, Union
from spacy.util import registry
import numpy
import tempfile

import nmslib
import scipy
from sklearn.feature_extraction.text import TfidfVectorizer
from crtl.lkb.lkb import LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path


class MentionCandidate(NamedTuple):
    """
    A data class representing a candidate entity that a mention may be linked to.
    Parameters
    ----------
    concept_id : str, required.
        The canonical concept id in the KB.
    terms : List[str], required.
        The aliases that caused this entity to be linked.
    similarities : List[float], required.
        The cosine similarities from the mention text to the alias in tf-idf space.
    """

    concept_id: str
    similarity: float


class ANNCandidateGenerator:
    def __init__(
        self,
        lkb: LexicalKnowledgeBase,
        term_index: list[str],
        ann_index,  #: FloatIndex,
        term_vectors,
        vectorizer: TfidfVectorizer,
    ):
        self.lkb = lkb
        self._term_index = term_index
        self._ann_index = ann_index
        self._vectorizer = vectorizer
        self._vectors = term_vectors

    @classmethod
    def _create_vectorizer(cls, lkb, min_df) -> TfidfVectorizer:
        tfidf_vectorizer = TfidfVectorizer(
            analyzer="char_wb", ngram_range=(3, 3), min_df=min_df, dtype=numpy.float64
        )
        tfidf_vectorizer.fit_transform(lkb.term_to_cuis.keys())
        return tfidf_vectorizer

    @classmethod
    def initialize_candidate_generator(cls, lkb, min_df=10):
        terms = sorted(list(lkb.term_to_cuis.keys()))
        start_time = datetime.datetime.now()
        print(f"Creating  vectors...")
        vectorizer = cls._create_vectorizer(lkb, min_df)
        vectors = vectorizer.transform(terms)
        print(f"Created vectors.")
        print("Finding empty (all zeros) tfidf vectors")
        vectors, terms = cls._delete_empty_vectors(vectors, terms)
        ann_index = cls._create_ann_index(vectors)
        end_time = datetime.datetime.now()
        elapsed_time = end_time - start_time
        print(
            f"Creating candidate generator took {elapsed_time.total_seconds()} seconds"
        )
        return cls(lkb, terms, ann_index, vectors, vectorizer)

    @classmethod
    def _create_ann_index(cls, vectors):
        # nmslib hyperparameters (very important)
        # guide: https://github.com/nmslib/nmslib/blob/master/python_bindings/parameters.md
        # Default values resulted in very low recall.

        # set to the maximum recommended value. Improves recall at the expense of longer indexing time.
        # We use the HNSW (Hierarchical Navigable Small World Graph) representation which is constructed
        # by consecutive insertion of elements in a random order by connecting them to M closest neighbours
        # from the previously inserted elements. These later become bridges between the network hubs that
        # improve overall graph connectivity. (bigger M -> higher recall, slower creation)
        # For more details see:  https://arxiv.org/pdf/1603.09320.pdf?
        m_parameter = 100
        # `C` for Construction. Set to the maximum recommended value
        # Improves recall at the expense of longer indexing time
        construction = 2000
        num_threads = 60  # set based on the machine
        index_params = {
            "M": m_parameter,
            "indexThreadQty": num_threads,
            "efConstruction": construction,
            "post": 0,
        }

        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(vectors)
        ann_index.createIndex(index_params, print_progress=True)
        return ann_index

    @classmethod
    def _delete_empty_vectors(cls, vectors, terms):
        empty_tfidfs_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        number_of_non_empty_tfidfs = sum(
            empty_tfidfs_boolean_flags == False
        )  # noqa: E712
        total_number_of_tfidfs = numpy.size(vectors, 0)
        print(
            f"Deleting {number_of_non_empty_tfidfs}/{total_number_of_tfidfs} aliases because their tfidf is empty"
        )
        # remove empty tfidf vectors, otherwise nmslib will crash
        terms = [
            alias for alias, flag in zip(terms, empty_tfidfs_boolean_flags) if flag
        ]
        vectors = vectors[empty_tfidfs_boolean_flags]
        assert len(terms) == numpy.size(vectors, 0)
        return vectors, terms

    def nmslib_knn_with_zero_vectors(
        self, vectors: numpy.ndarray, k: int
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        ann_index.knnQueryBatch crashes if any of the vectors is all zeros.
        This function is a wrapper around `ann_index.knnQueryBatch` that solves this problem. It works as follows:
        - remove empty vectors from `vectors`.
        - call `ann_index.knnQueryBatch` with the non-empty vectors only. This returns `neighbors`,
        a list of list of neighbors. `len(neighbors)` equals the length of the non-empty vectors.
        - extend the list `neighbors` with `None`s in place of empty vectors.
        - return the extended list of neighbors and distances.
        """
        empty_vectors_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        empty_vectors_count = vectors.shape[0] - sum(empty_vectors_boolean_flags)

        # init extended_neighbors with a list of Nones
        extended_neighbors = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )
        extended_distances = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )

        if vectors.shape[0] - empty_vectors_count == 0:
            return extended_neighbors, extended_distances

        # remove empty vectors before calling `ann_index.knnQueryBatch`
        vectors = vectors[empty_vectors_boolean_flags]

        # call `knnQueryBatch` to get neighbors
        original_neighbours = self._ann_index.knnQueryBatch(vectors, k=k)

        neighbors, distances = zip(
            *[(x[0].tolist(), x[1].tolist()) for x in original_neighbours]
        )
        neighbors = list(neighbors)
        distances = list(distances)

        # neighbors need to be converted to an np.array of objects instead of ndarray of dimensions len(vectors)xk
        # Solution: add a row to `neighbors` with any length other than k. This way, calling np.array(neighbors)
        # returns an np.array of objects
        neighbors.append([])
        distances.append([])
        # interleave `neighbors` and Nones in `extended_neighbors`
        extended_neighbors[empty_vectors_boolean_flags] = numpy.array(
            neighbors, dtype=object
        )[:-1]
        extended_distances[empty_vectors_boolean_flags] = numpy.array(
            distances, dtype=object
        )[:-1]

        return extended_neighbors, extended_distances

    def to_disk(self, path: Union[pathlib.Path, str]):
        path = pathlib.Path(path)
        path.mkdir(exist_ok=True)
        scipy.sparse.save_npz(path / "vectors.npz", self._vectors.astype(numpy.float16))
        self.lkb.to_disk(path / "lkb")
        self._ann_index.saveIndex(str(path / "ann_index"))
        with open(path / "term_index", "w") as f:
            for term in self._term_index:
                f.write(term + "\n")
        with open(path / "vectorizer_params.json", "w") as f:
            json.dump({"min_df": self._vectorizer.get_params()["min_df"]}, f)

    @classmethod
    def from_disk(cls, path: pathlib.Path):
        term_vectors = scipy.sparse.load_npz(path / "vectors.npz").astype(numpy.float32)
        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(term_vectors)
        ann_index.loadIndex(str(path / "ann_index"))
        lkb = LexicalKnowledgeBase.from_disk(path / "lkb")
        term_index = [term.strip() for term in open(path / "term_index").readlines()]
        min_df = json.load(open(path / "vectorizer_params.json"))["min_df"]
        vectorizer = cls._create_vectorizer(lkb, min_df)
        return cls(lkb, term_index, ann_index, term_vectors, vectorizer)

    def set_runtime_parameters(self, runtime_params: Dict):
        ef_search = runtime_params["ef_search"]
        if ef_search:
            self._ann_index.setQueryTimeParams({"efSearch": ef_search})

    def __call__(
        self, terms: Iterable[str], k: int
    ) -> Iterable[Iterable[MentionCandidate]]:
        # tfidf vectorizer crashes on an empty array, so we return early here
        terms = list(terms)
        if len(terms) == 0:
            return []
        vectors = self._vectorizer.transform(terms)
        # `ann_index.knnQueryBatch` crashes if one of the vectors is all zeros.
        # `nmslib_knn_with_zero_vectors` is a wrapper around `ann_index.knnQueryBatch` that addresses this issue.
        batch_mention_candidates = self._get_batch_mention_candidates(vectors, k)
        return batch_mention_candidates

    def _get_batch_mention_candidates(self, vectors: numpy.ndarray, k: int):
        batch_neighbors, batch_distances = self.nmslib_knn_with_zero_vectors(vectors, k)
        batch_mention_candidates = []
        for neighbors, distances in zip(batch_neighbors, batch_distances):
            mention_candidates = self._get_mention_candidates(neighbors, distances)
            batch_mention_candidates.append(mention_candidates)
        return batch_mention_candidates

    def _get_mention_candidates(self, neighbors, distances):
        if neighbors is None:
            neighbors = []
        if distances is None:
            distances = []
        concept_to_mentions: Dict[str, List[str]] = defaultdict(list)
        concept_to_similarities: Dict[str, List[float]] = defaultdict(list)
        for neighbor_index, distance in zip(neighbors, distances):
            mention = self._term_index[neighbor_index]
            concepts_for_mention = self.lkb.term_to_cuis[mention]
            for concept_id in concepts_for_mention:
                concept_to_mentions[concept_id].append(mention)
                concept_to_similarities[concept_id].append(1.0 - distance)

        mention_candidates = [
            MentionCandidate(concept, max(concept_to_similarities[concept]))
            for concept in concept_to_mentions.keys()
        ]
        mention_candidates = sorted(
            mention_candidates, key=lambda c: c.similarity, reverse=True
        )
        return mention_candidates

    @classmethod
    def from_gzip(cls, path: pathlib.Path):
        with tempfile.TemporaryDirectory() as tmpdirname:
            tar = tarfile.open(path)
            file_name = tar.getnames()[0]
            tar.extractall(tmpdirname)
            return cls.from_disk(pathlib.Path(tmpdirname) / file_name)

@registry.misc("load_ann_candidate_generator")
def load_ann_candidate_generator(url_str_or_path: Union[str, pathlib.Path]):
    path: pathlib.Path = cached_path(url_str_or_path)
    if path.suffix == ".gz":
        cg = ANNCandidateGenerator.from_gzip(path)
    else:
        cg = ANNCandidateGenerator.from_disk(path)
    return lambda: cg


# Should an abstract candidate generator be passing mentioncandidates that contain reference
# to the terms they matched on, and their respective similarities? If an candidate generator is really supposed to be
# returning conceptual candidates, returning this information seems like an implementation detail.
# Mention Candidates no longer return term matches, only concept id and its similarity to the span.


